use num::FromPrimitive;
// Used FromPrimitive from num

enum_from_primitive!{
	pub enum Instruction {
		PUSH    = 0x00,
		POP     = 0x01,
		ADD     = 0x02,
		SUBS    = 0x03,
		MULT    = 0x04,
		DIV     = 0x05,
		REMN    = 0x06,
		INTOUT  = 0x07,
		CHAROUT = 0x08,
		STROUT  = 0x09,
		NEWLINE = 0x0A,
		MEMFREE = 0x0B,
		VER     = 0x0C,
		PANIC   = 0x0D,
		THROW   = 0x0E,
		EXIT    = 0x0F
	}
}

pub fn jBrain_version() -> &'static str {
	return "0.1.0";
}

pub struct jBrainVM{
	push: bool,
	stack: Vec<u8>
}

impl jBrainVM{
	pub fn new() -> jBrainVM{
		println!("Starting jBrain instance...");
		jBrainVM{push: false, stack: vec![]}
	}
	pub fn interpreter(&mut self, bytecode: String) -> (){
		for execbyte in bytecode.chars() {
			self.execute(execbyte as u8);
		}
	}
	fn execute(&mut self, execbyte: u8) -> () {
		if self.push {
			self.push(execbyte);
            self.push=false;

		} else{
			let op: Option<Instruction> = Instruction::from_u8(execbyte);
			match op{
				None => {
					panic!("Unknown instruction, ERROR!!");
				},
				Some(bc) => {
					match bc{
						Instruction::PUSH => {
							self.push = true;
						},
						Instruction::POP => {
							self.pop();
						},
						Instruction::ADD => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;
							let c = a + b;
							self.push(c as u8);
						},
						Instruction::SUBS => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;
							let c = a - b;
							self.push(c as u8);
						},
						Instruction::MULT => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;
							let c = a * b;
							self.push(c as u8);
						},
						Instruction::DIV => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;
							let c = a / b;
							self.push(c as u8);
						},
						Instruction::REMN => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;
							let c = a % b;
							self.push(c as u8);
						},
						Instruction::INTOUT => {
							print!("{}", self.pop() as i32);
						},
						Instruction::CHAROUT => {
							print!("{}", self.pop() as char);
						},
						Instruction::STROUT => {
							let mut output = String::new();
							for i in 0..self.stack.len() {
								output = format!("{}{}", output, self.stack[i] as char);
							}
							print!("{}", output);
						},
						Instruction::NEWLINE => {
							print!("\n");
						},
						Instruction::MEMFREE => {
							self.stack = vec![];
						},
						Instruction::VER => {
							println!("jBrainVM v{}", jBrain_version());
						},
						Instruction::PANIC => {
							panic!("This program crashed with ERR(01)");
						},
						Instruction::THROW => {
							std::process::exit(self.pop() as i32);
						},
						Instruction::EXIT => {
							println!("\n[Program finished successfully]");
							std::process::exit(0);
						}
					}
				}
			}
			
		}
    }
    
	fn push(&mut self, value: u8) -> (){
		self.stack.push(value);
	}
	fn pop(&mut self) -> u8{
		let a: Option<u8>=self.stack.pop();
		match a{
			None => {
				println!("Failed to pop");
				0
			},
			Some(result) => {
				result
			}
		}
	}
}